export class Result {
  #value
  #error

  constructor(value, error) {
    this.#value = value
    this.#error = error
  }

  static Ok(value) {
    return new Result(value, null)
  }

  static Err(error) {
    return new Result(null, error)
  }

  isOk() {
    return this.#error === null
  }

  isErr() {
    return !this.isOk()
  }

  map(fn) {
    if (this.isOk()) {
      return Result.Ok(fn(this.#value))
    } else {
      return this
    }
  }

  flatMap(fn) {
    if (this.isOk()) {
      return fn(this.#value)
    } else {
      return this
    }
  }

  unwrap() {
    if (this.isOk()) {
      return this.#value
    } else {
      throw new ResultError(`failed to unwrap result: ${this.#error.message}`, this)
    }
  }

  unwrapOr(value) {
    return this.isOk() ? this.#value : value
  }
}

export class ResultError extends TypeError {
  constructor(msg, result) {
    super(msg)
    this.result = result
  }
}

